@@ -1,4 +1,4 @@
-# $Id: Client-Keepalive.pm 99 2008-12-07 03:52:29Z rcaputo $
+# $Id: Client-Keepalive.pm 111 2009-07-28 06:06:04Z rcaputo $
 
 package POE::Component::Client::Keepalive;
 
@@ -6,16 +6,17 @@ use warnings;
 use strict;
 
 use vars qw($VERSION);
-$VERSION = "0.24";
+$VERSION = "0.260";
 
 use Carp qw(croak);
-use Errno qw(ETIMEDOUT);
+use Errno qw(ETIMEDOUT EBADF);
 use Socket qw(SOL_SOCKET SO_LINGER);
 
 use POE;
 use POE::Wheel::SocketFactory;
 use POE::Component::Connection::Keepalive;
 use POE::Component::Client::DNS;
+use Net::IP qw(ip_is_ipv4);
 
 my $ssl_available;
 eval {
@@ -156,20 +157,21 @@ sub new {
       $self => {
         _start               => "_ka_initialize",
         _stop                => "_ka_stopped",
+        ka_add_to_queue      => "_ka_add_to_queue",
+        ka_cancel_dns_response => "_ka_cancel_dns_response",
         ka_conn_failure      => "_ka_conn_failure",
         ka_conn_success      => "_ka_conn_success",
+        ka_deallocate        => "_ka_deallocate",
+        ka_dns_response      => "_ka_dns_response",
+        ka_keepalive_timeout => "_ka_keepalive_timeout",
         ka_reclaim_socket    => "_ka_reclaim_socket",
         ka_relinquish_socket => "_ka_relinquish_socket",
         ka_request_timeout   => "_ka_request_timeout",
+        ka_resolve_request   => "_ka_resolve_request",
         ka_set_timeout       => "_ka_set_timeout",
         ka_shutdown          => "_ka_shutdown",
         ka_socket_activity   => "_ka_socket_activity",
-        ka_keepalive_timeout => "_ka_keepalive_timeout",
         ka_wake_up           => "_ka_wake_up",
-        ka_resolve_request   => "_ka_resolve_request",
-        ka_add_to_queue      => "_ka_add_to_queue",
-        ka_dns_response      => "_ka_dns_response",
-        ka_cancel_dns_response => "_ka_cancel_dns_response",
       },
     ],
   );
@@ -254,6 +256,7 @@ sub _ka_wake_up {
 
       # Remove the wheel-to-request index.
       delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+      _free_req_id($request->[RQ_ID]);
 
       next;
     }
@@ -277,6 +280,7 @@ sub _ka_wake_up {
     # Move the wheel and its request into SF_WHEELS.
     DEBUG and warn "WAKEUP: creating wheel for $req_key";
 
+    # TODO - Set the SocketDomain to AF_INET6 if $addr =~ /:/?
     my $addr = ($request->[RQ_IP] or $request->[RQ_ADDRESS]);
     my $wheel = POE::Wheel::SocketFactory->new(
       BindAddress   => $self->[SF_BIND_ADDR],
@@ -416,48 +420,54 @@ sub deallocate {
     defined($req_id) and exists($active_req_ids{$req_id})
   );
 
-  my $request = $self->[SF_REQ_INDEX]{$req_id};
+  my $request = delete $self->[SF_REQ_INDEX]{$req_id};
   unless (defined $request) {
     DEBUG_DEALLOCATE and warn "deallocate could not find request $req_id";
     return;
   }
+  _free_req_id($request->[RQ_ID]);
+
+  # Now pass the vetted request & its ID into our manager session.
+  $poe_kernel->call("$self", "ka_deallocate", $request, $req_id);
+}
+
+sub _ka_deallocate {
+  my ($self, $heap, $request, $req_id) = @_[OBJECT, HEAP, ARG0, ARG1];
 
   my $conn_key = $request->[RQ_CONN_KEY];
   my $existing_connection = $self->_check_free_pool($conn_key);
 
+  # Existing connection.  Remove it from the pool, and delete the socket.
   if (defined $existing_connection) {
-    # remove it from the pool, delete the socket
     $self->_remove_socket_from_pool($existing_connection->{socket});
     DEBUG_DEALLOCATE and warn(
       "deallocate called, deleted already-connected socket"
     );
     return;
   }
-  else {
+
+  # No connection yet.  Cancel the request.
+  DEBUG_DEALLOCATE and warn(
+    "deallocate called without an existing connection.  ",
+    "cancelling connection request"
+  );
+
+  unless (exists $heap->{resolve}->{$request->[RQ_ADDRESS]}) {
     DEBUG_DEALLOCATE and warn(
-      "deallocate called without an existing connection.  ",
-      "cancelling connection request"
+      "deallocate cannot cancel dns -- no pending request"
     );
-
-    my $heap = $poe_kernel->get_active_session->get_heap;
-
-    unless (exists $heap->{resolve}->{$request->[RQ_ADDRESS]}) {
-      DEBUG_DEALLOCATE and warn(
-        "deallocate cannot cancel dns -- no pending request"
-      );
-      return;
-    }
-
-    if ($heap->{resolve}->{$request->[RQ_ADDRESS]} eq 'cancelled') {
-      DEBUG_DEALLOCATE and warn(
-        "deallocate cannot cancel dns -- request already cancelled"
-      );
-      return;
-    }
-
-    $poe_kernel->call( "$self", ka_cancel_dns_response => $request );
     return;
   }
+
+  if ($heap->{resolve}->{$request->[RQ_ADDRESS]} eq 'cancelled') {
+    DEBUG_DEALLOCATE and warn(
+      "deallocate cannot cancel dns -- request already cancelled"
+    );
+    return;
+  }
+
+  $poe_kernel->call( "$self", ka_cancel_dns_response => $request );
+  return;
 }
 
 sub _ka_cancel_dns_response {
@@ -542,6 +552,7 @@ sub _ka_conn_failure {
 
   # remove the wheel-to-request index
   delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+  _free_req_id($request->[RQ_ID]);
 
   # Discount the use by request key, removing the SF_USED record
   # entirely if it's now moot.
@@ -564,6 +575,7 @@ sub _ka_conn_success {
 
   # remove the wheel-to-request index
   delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+  _free_req_id($request->[RQ_ID]);
 
   # Remove the SF_USED placeholder, add in the socket, and store it
   # properly.
@@ -706,15 +718,18 @@ sub _ka_reclaim_socket {
   if ($socket_is_active) {
     DEBUG and warn "RECLAIM: socket is still active; trying to drain";
     use bytes;
+
     my $socket_had_data = sysread($socket, my $buf = "", 65536) || 0;
     DEBUG and warn "RECLAIM: socket had $socket_had_data bytes. 0 means EOF";
     DEBUG and warn "RECLAIM: Giving up on socket.";
 
-    # Avoid common FIN_WAIT_2 issues.
+    # Avoid common FIN_WAIT_2 issues, but only for valid sockets.
+    #if ($socket_had_data and fileno($socket)) {
     if ($socket_had_data) {
-      setsockopt($socket, SOL_SOCKET, SO_LINGER, pack("sll",1,0,0)) or die(
-        "setsockopt: $!"
+      my $opt_result = setsockopt(
+        $socket, SOL_SOCKET, SO_LINGER, pack("sll",1,0,0)
       );
+      die "setsockopt: " . ($!+0) . " $!" if (not $opt_result and $!  != EBADF);
     }
 
     goto &_ka_wake_up;
@@ -786,6 +801,10 @@ sub _ka_shutdown {
 
   # Stop any pending resolver requests.
   foreach my $host (keys %{$heap->{resolve}}) {
+    if ($heap->{resolve}{$host} eq 'cancelled') {
+      DEBUG and warn "SHT: Skipping shutdown for $host (already cancelled)";
+      next;
+    }
     DEBUG and warn "SHT: Shutting down resolver requests for $host";
     foreach my $request (@{$heap->{resolve}{$host}}) {
       $self->_shutdown_request($kernel, $request);
@@ -819,6 +838,7 @@ sub _shutdown_request {
 
     # remove the wheel-to-request index
     delete $self->[SF_REQ_INDEX]{$request->[RQ_ID]};
+    _free_req_id($request->[RQ_ID]);
   }
 
   if (defined $request->[RQ_SESSION]) {
@@ -857,8 +877,11 @@ sub _ka_resolve_request {
   my $host = $request->[RQ_ADDRESS];
 
   # Skip DNS resolution if it's already a dotted quad.
-  # TODO - Not all dotted quads are good.
-  if ($host =~ /^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$/) {
+  # ip_is_ipv4() doesn't require quads, so we count the dots.
+  #
+  # TODO - Do the same for IPv6 addresses containing colons?
+  # TODO - Would require AF_INET6 support around the SocketFactory.
+  if ((($host =~ tr[.][.]) == 3) and ip_is_ipv4($host)) {
     DEBUG_DNS and warn "DNS: $host is a dotted quad; skipping lookup";
     $kernel->call("$self", ka_add_to_queue => $request);
     return;
@@ -1326,15 +1349,9 @@ keep-alive pool, allowing a program to terminate gracefully.
 L<POE>
 L<POE::Component::Connection::Keepalive>
 
-=head1 BUGS
-
-http://rt.cpan.org/NoAuth/Bugs.html?Dist=POE-Component-Client-Keepalive
-tracks the known issues with this component.  You can add to them by
-sending mail to bug-poe-component-client-keepalive@rt.cpan.org.
-
 =head1 LICENSE
 
-This distribution is copyright 2004-2006 by Rocco Caputo.  All rights
+This distribution is copyright 2004-2009 by Rocco Caputo.  All rights
 are reserved.  This distribution is free software; you may
 redistribute it and/or modify it under the same terms as Perl itself.
 
@@ -1349,4 +1366,16 @@ Rob Bloodgood helped out a lot.  Thank you.
 Joel Bernstein solved some nasty race conditions.  Portugal Telecom
 L<http://www.sapo.pt/> was kind enough to support his contributions.
 
+=head1 BUG TRACKER
+
+https://rt.cpan.org/Dist/Display.html?Status=Active&Queue=POE-Component-Client-Keepalive
+
+=head1 REPOSITORY
+
+http://thirdlobe.com/svn/poco-client-keepalive/
+
+=head1 OTHER RESOURCES
+
+http://search.cpan.org/dist/POE-Component-Client-Keepalive/
+
 =cut
